from re import search
from typing import List

import matplotlib.pyplot as plt
import numpy as np
import time
import sys



###### PARTIE 2 ######

def experiment_load_factor(load_factors):
    """
    Etude l'inmpact du facteur de charge sur les performances d'un dictionnaire Python.
    Mesure les temps d'insertion, de recherche et de suppression pour différents facteurs de charge.
    Returns:
        insetion_times: liste contenant les temps d'insertion (en secondes) pour chaque facteur de charge.
        search_times: Liste contenant les temps de recherchde (en secondes) pour chaque facteur de charge.
        deletion_times: Liste contenant les temps de suppression (en secondes) pour chaque facteur de charge
    """
    insertion_times = []  # Pour stocker le temps total d'insertion
    num_resizes = []  # Pour stocker le nombre de réallocations
    sizes = []  # Pour stocker la taille mémoire finale

    for lf in load_factors:
        # Crée un dictionnaire vide
        d = {}

        # Nombre d'éléments à insérer pour ce facteur de charge
        num_elements = int(10000 * lf)

        # Variables pour suivre la taille courante et détecter une réallocation
        last_size = sys.getsizeof(d)  # taille mémoire initiale
        resize_count = 0

        # Mesure du temps d'insertion (global)
        start_time = time.time()

        for i in range(num_elements):
            d[i] = i

            # Après chaque insertion, vérifie si la taille mémoire a augmenté
            current_size = sys.getsizeof(d)
            if current_size > last_size:
                resize_count += 1
                last_size = current_size

        end_time = time.time()
        total_insertion_time = end_time - start_time

        # Stockage des résultats pour ce facteur de charge
        insertion_times.append(total_insertion_time)
        num_resizes.append(resize_count)
        sizes.append(sys.getsizeof(d))

    return insertion_times, num_resizes, sizes



def experiment_longest():
    """
    Effectue 10 000 insertions successives dans un dict, en mesurant
    le temps d'insertion pour chacune.
    Retourne l'histogramme (fréquences) de ces temps d'insertion,
    discrétisés en bacs (bins).
    """
    d = {}
    insertion_times = []

    for i in range(10000):
        key = str(i)
        value = i
        start_time = time.time()
        d[key] = value
        insertion_time = time.time() - start_time
        insertion_times.append(insertion_time)
    frequencies = np.histogram(insertion_times)[0]
    return frequencies

def visualisation(load_factors, insertion_times, num_resizes, sizes, frequencies):
    """
    Visualisation des résultats
    """
    # Temps d'insertion en fonction du facteur de charge
    plt.figure(figsize=(6, 4))
    plt.plot(load_factors, insertion_times, marker='o')
    plt.xlabel('Facteur de charge (load_factor)')
    plt.ylabel('Temps total d\'insertion (secondes)')
    plt.title('Temps d\'insertion en fonction du facteur de charge')
    plt.grid(True)
    plt.tight_layout()
    plt.savefig('Tps_insertion.facteur_de_charge.png')
    plt.close()

    # Nombre de réallocations de mémoire en fonction du facteur de charge
    plt.figure(figsize=(6, 4))
    plt.plot(load_factors, num_resizes, marker='o')
    plt.xlabel('Facteur de charge (load_factor)')
    plt.ylabel('Nombre de réallocations (resizes)')
    plt.title('Nombre de réallocations de mémoire en fonction du facteur de charge')
    plt.grid(True)
    plt.tight_layout()
    plt.savefig('nb_reallocation.facteur_de_charge.png')
    plt.close()

    # Taille de mémoire occupée en fonction du nombre d'éléments
    #    Le nombre d'éléments insérés = 10000 * load_factor
    nb_elements = [int(10000 * lf) for lf in load_factors]
    plt.figure(figsize=(6, 4))
    plt.plot(nb_elements, sizes, marker='o')
    plt.xlabel('Nombre d\'éléments insérés')
    plt.ylabel('Taille mémoire du dict (octets)')
    plt.title('Taille de mémoire occupée en fonction du nombre d\'éléments')
    plt.grid(True)
    plt.tight_layout()
    plt.savefig('courbe_memoire.nombre_delement.png')
    plt.close()

    # Deuxième étude

    f = frequencies
    plt.figure(figsize=(10, 6))
    plt.bar(range(len(f)), f)
    plt.xlabel('Temps d\'insertion (s)')
    plt.ylabel('Fréquence')
    plt.title('Histogramme des fréquences des temps d\'insertions')
    plt.yscale('log')
    xticks = np.logspace(-6, 1, 3)
    xtick_labels = [f'{x:.1e}' for x in xticks]
    plt.xticks(xticks, xtick_labels)
    plt.savefig('histogramme.png')


load_factors = [0.01, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
insertion_times, num_resizes, sizes = experiment_load_factor(load_factors)
frequencies = experiment_longest()
visualisation(load_factors, insertion_times, num_resizes, sizes, frequencies)
