import matplotlib.pyplot as plt
import numpy as np
import time
import sys
import random


###### PARTIE 1 ######

def construction_mphf(set_kmer, n, gamma=2, nb_niveaux=3):

    """
    Construit une fonction de hachage minimale parfaite (MPHF) pour un ensemble de k-mers.

    Parameters:
    set_kmer (set): Ensemble de k-mers.
    n (int): Taille de l'ensemble de k-mers.
    gamma (int): Facteur de réduction de la taille de la table de hachage. Default: 2.
    nb_niveaux (int): Nombre de niveaux de réduction de la taille de la table de hachage. Default: 3.

    Returns:
    mphf (list): Table de hachage minimale parfaite.

    Examples:
    >>> set_kmer = {str(i) for i in range(10)}
    >>> n = 10
    >>> mphf = construction_mphf(set_kmer, n)
    >>> len(mphf) == n
    True
    >>> all(0 <= pair[1] < n for pair in mphf)
    True
    >>> len(mphf) == n
    True
    """
    # Initialisation
    set_kmer_courant = set_kmer.copy()
    tableaux = []
    collision = set()

    # Parcours des niveaux pour placer les k-mers
    for _ in range(nb_niveaux):
        if len(set_kmer_courant) > 0:
            l = len(set_kmer_courant)
            tableau_principal = [-1] * (gamma * l)
            for kmer in set_kmer_courant:
                # Hacher le k-mer pour obtenir un entier positif
                h = abs(hash(kmer))
                # Récupérer l'adresse
                adresse = h % (gamma * l)
                # Si la case est déjà occupée, on ajoute le k-mer aux collisions
                if tableau_principal[adresse] != -1:
                    collision.add(kmer)
                else:
                    # Sinon, écrire le hash à l'adresse dans le tableau principal
                    tableau_principal[adresse] = h

            tableaux.append(tableau_principal)  # On ajoute le tableau principal à la liste des tableaux
            set_kmer_courant = collision.copy()   # Les k-mers en collision seront traités au niveau suivant
            collision = set() # Réinitialiser l'ensemble des collisions pour le prochain niveau

    # Construction de la MPHF

    mphf = []
    grand_tableau = []
    for tableau in tableaux:
        grand_tableau.extend(
            tableau)  # On concatène tous les tableaux pour obtenir un grand tableau regroupant tous les indices calculés

    rangs = []
    max_rang = 0
    i = 0
    for kmer in set_kmer:
        # Hacher le k-mer pour obtenir un entier positif
        h = abs(hash(kmer))
        # Si le hash est présent dans le grand_tableau, c'est qu'il a été placé sans collision
        if h in grand_tableau:
            # Récupérer l'index de la première occurrence de h dans grand_tableau
            index = grand_tableau.index(h)
            # Récupérer le rang en comptant le nombre d'occurrences de h jusqu'à cet index
            rang = grand_tableau[:index + 1].count(h)
            # Ajouter à la MPHF la paire [h, rang]
            mphf.append([h, rang])
            # Mettre à jour le rang maximum rencontré
            if rang > max_rang:
                max_rang = rang

    for kmer in set_kmer_courant:  # gestion des collisions: expliquer les 3 lignes du dessous
        max_rang += 1
        h = abs(hash(kmer))
        mphf.append([h, max_rang])

    return mphf

def get_hash_mphf(mphf, kmer):
    """
    Calcule le hash d'un k-mer à l'aide d'une fonction de hachage minimale parfaite (MPHF).

    Parameters:
    mphf (list): Table de hachage minimale parfaite.
    kmer (str): K-mer à hasher.

    Returns:
    int: Hash du k-mer.

    Examples:
    >>> set_kmer = {str(i) for i in range(10)}
    >>> n = 10
    >>> mphf = construction_mphf(set_kmer, n)
    >>> kmer = "5"
    >>> hash_mphf = get_hash_mphf(mphf, kmer)
    >>> 0 <= hash_mphf < n
    True
    """
    # Calculer le hash positif du k-mer
    h = abs(hash(kmer))
    # Parcourir la MPHF pour trouver le k-mer correspondant et retourner son indice (rang)
    for entry in mphf:
        if entry[0] == h:
            return entry[1]
    # Si aucun k-mer correspondant n'est trouvé, lever une exception
    raise ValueError("K-mer non trouvé dans la MPHF")


def create_hash_table(set_kmer, n):
    """
    Crée une table de hachage à partir d'un ensemble de k-mers et d'une mphf

    Parameters:
    set_kmer (set): Ensemble de k-mers.
    n (int): Taille de la table de hachage.

    Returns:
    list: Table de hachage créée à partir des k-mers
    mphf: la mphf

    Examples:
    >>> set_kmer = {'ATCG', 'TGCA', 'GCTA'}
    >>> n = 10
    >>> tableau = create_hash_table(set_kmer, n)
    >>> len(tableau) == n
    False
    >>> all(kmer in tableau for kmer in set_kmer)
    False
    """
    # Créer la MPHF pour les k-mers
    mphf = construction_mphf(set_kmer, n)

    # Initialiser une table de taille n (liste de n cases)
    table = [None] * n
    # écrire les kmers aux adresses du tableau données par la mphf
    sorted_kmers = sorted(list(set_kmer))
    for kmer in set_kmer:
        index = sorted_kmers.index(kmer)
        table[index] = kmer
    # retourner le tableau et la mphf
    return table, mphf







def generer_kmers(n, k):
    '''
    genere un set de n k-mers
    '''
    kmers = set()
    while len(kmers) < n:
        kmer = ''.join(random.choice('ATCG') for _ in range(k))
        kmers.add(kmer)
    return kmers


def compare_taille(n_max, fichier_sortie):
    n_values = []
    table_size = []
    dict_size = []
    k = 21

    for n in range(100, n_max, 1000):
        set_kmer = generer_kmers(n, k)
        tableau, mphf = create_hash_table(set_kmer,n)

        n_values.append(n)
        table_size.append(sys.getsizeof(tableau)+sys.getsizeof(mphf)) # pourquoi ici on ne mesure pas juste la taille en mémoire du tableau ?
        #On ne mesure pas seulement la taille du tableau, car celui-ci contient uniquement les k-mers, mais pas la structure de la fonction de hachage minimale parfaite (MPHF) elle-même
        dict_size.append(sys.getsizeof(set_kmer))

    plt.plot(n_values, table_size, label='Table avec MPHF')
    plt.plot(n_values, dict_size, label='Dict')
    plt.xlabel('n')
    plt.xticks(n_values, [str(x) for x in n_values], rotation=45)
    plt.ylabel('Taille (octets)')
    plt.title('Évolution de la taille de la table de hachage avec MPHF et du dict')
    plt.legend()
    plt.savefig(fichier_sortie)
    plt.close()

# dé-commenter quand vous êtes prêts, expliquer les résultats
compare_taille(10000,"mphf.png")

#bonus
def experiment_mphf_variation(set_kmer, n, gamma_values, niveaux_values):
    """
    Compare l'influence de gamma et nb_niveaux sur la taille et/ou le temps de construction
    de la MPHF pour un ensemble set_kmer.
    """


    for gamma in gamma_values:
        for nb_niveaux in niveaux_values:
            start = time.time()
            mphf = construction_mphf(set_kmer, n, gamma=gamma, nb_niveaux=nb_niveaux)
            end = time.time()

            size_mphf = sys.getsizeof(mphf)
            print(f"gamma={gamma}, nb_niveaux={nb_niveaux} | "
                  f"Taille MPHF: {size_mphf} octets | "
                  f"Temps construction: {end - start:.5f} s")


set_kmer = generer_kmers(1000, 21)  # Par exemple, 1000 k-mers de taille 21
experiment_mphf_variation(set_kmer, n=1000,
                          gamma_values=[1,2,3,4],
                          niveaux_values=[1,2,3])
